#ifndef STREU_CPLUSLIB_SMARTPTR_H_INCLUDED
#define STREU_CPLUSLIB_SMARTPTR_H_INCLUDED

#include "cpluslib/config.h"
#include "cpluslib/object.h"
#include <cstdlib>

#ifdef STREU_CONFIG_DEBUG
void cpluslibEnsurePtrValid();
#else
#  define cpluslibEnsurePtrValid()
#endif

/* reference-counted pointer */
template<class T>
class Ptr {
 public:
    /* This one's exported for various reasons. In particular,
       Ptr<T>::Ref is POD, so it can be put in a union. Just be
       sure to update the reference counts accordingly... */
    struct Ref {
        T* ptr;
        ReferenceCount* rc;
    };
    Ref ref;
 //protected:
    /* decrement reference count */
    void dec() const {
        //if (ref.ptr) cpluslibEnsurePtrValid();
        if(!--ref.rc->ptr_count) {
            delete ref.ptr;
        }
    };
    /* increment reference count */
    void inc() const {
        if (ref.ptr) cpluslibEnsurePtrValid();
        ++ref.rc->ptr_count;
    }
 public:
    /* ctors */
    Ptr() {
        ref.ptr = 0;
        ref.rc = &ReferenceCount::null_ref;
        ++ref.rc->ptr_count;
    }
    Ptr(T* ap) {
        ref.ptr = ap;
        ref.rc = ReferenceCount::MapPointer(ap);
        inc();
    }

    Ptr(const Ptr& a) : ref(a.ref) {
        inc();
    }

    template<class T1>
    Ptr(const Ptr<T1>& a) {
        ref.ptr = a.ref.ptr;
        ref.rc  = a.ref.rc;
        inc();
    }

    /* dtor */
    ~Ptr() { dec(); };

    /* copy operator */
    Ptr& operator=(const Ptr& a) {
        Ref r = a.ref;
        a.inc();
        dec();
        ref = r;
        return *this;
    }

    template<class T1>
    Ptr& operator=(const Ptr<T1>& a) {
        Ref r;
        r.ptr = a.ref.ptr;
        r.rc  = a.ref.rc;
        a.inc();
        dec();
        ref = r;
        return *this;
    }

    Ptr& operator=(T* a) {
        ReferenceCount* new_ref = ReferenceCount::MapPointer(a);
        ++new_ref->ptr_count;
        dec();
        ref.rc = new_ref;
        ref.ptr = a;
        return *this;
    }

    /* accessors */
    T* ptr() const { return ref.ptr; }
    std::size_t ref_count() const { return ref.rc->ptr_count; }

    operator T*const() const { return ref.ptr; }

    T& operator*() const { return *ref.ptr; }
    T* operator->() const { return ref.ptr; }

    /* dynamic_cast */
    template<class T1>
    Ptr<T1> cast() const { return dynamic_cast<T1*>(this->ptr()); }
};

/* reference-counted pointer to an array */
template<class T>
class ArrayPtr {
 public:
    struct Ref {
        T* ptr;
        ReferenceCount* rc;
        Ref(T* aptr, ReferenceCount* arc) : ptr(aptr), rc(arc) { }
        Ref(T* aptr) : ptr(aptr), rc(ReferenceCount::MapAPointer(aptr)) { }
    };
    Ref ref;
 //protected:
    /* decrement reference count */
    void dec() const {
        if(!--ref.rc->ptr_count) {
            delete[] ref.ptr;
        }
    };
    /* increment reference count */
    void inc() const { ++ref.rc->ptr_count; }
 public:
    /* ctors */
    ArrayPtr(T* ap = 0) : ref(ap, ReferenceCount::MapAPointer(ap)) {
        inc();
    }

    ArrayPtr(const ArrayPtr& a) : ref(a.ref) {
        inc();
    }

    ArrayPtr(std::size_t size) : ref(new T[size]) {
        inc();
    }

    /* dtor */
    ~ArrayPtr() { dec(); };

    /* copy operator */
    ArrayPtr& operator=(const ArrayPtr& a) {
        Ref r = a.ref;
        a.inc();
        dec();
        ref = r;
        return *this;
    }

    ArrayPtr& operator=(T* a) {
        ReferenceCount* new_ref = ReferenceCount::MapAPointer(a);
        ++new_ref->ptr_count;
        dec();
        ref.rc = new_ref;
        ref.ptr = a;
        return *this;
    }

    /* accessors */
    T* ptr() const { return ref.ptr; }
    std::size_t ref_count() const { return ref.rc->ptr_count; }

    T& operator*() const { return *ref.ptr; }
    T* operator->() const { return ref.ptr; }
    T& operator[](std::size_t ix) const { return ref.ptr[ix]; }
};

#endif
