/**
  *  \file substr.cc
  */

#include <cstring>
#include <cctype>
#include "cpluslib/substr.h"
#include "cpluslib/assert.h"
#include "cpluslib/cstring.h"

using namespace std;

#ifdef STREU_CONFIG_USE_STD_STRING
#  define data_or_null data
#endif

template<class T>
static inline T
xstr_min(T a, T b)
{
    return a < b ? a : b;
}

/** Construct from (static) C string. */
substring::substring(const charT* src)
    : str(src), len(strlen(src))
{ }

/** Construct from C++ string. */
substring::substring(const simple_string& str, size_type pos, size_type n)
{
    ASSERT(pos <= str.size());
    this->str = str.data_or_null() + pos;
    this->len = xstr_min(str.size() - pos, n);
}

substring::substring(const substring& str, size_type pos, size_type n)
{
    ASSERT(pos <= str.size());
    this->str = str.data_or_null() + pos;
    this->len = xstr_min(str.size() - pos, n);
}

substring::substring(const charT* src, size_type n)
    : str(src), len(n)
{ }

void
substring::swap(substring& rhs)
{
    substring tmp = *this;
    *this = rhs;
    rhs = tmp;
}

substring::size_type
substring::copy(charT* s, size_type n, size_type pos) const
{
    ASSERT(pos <= length());
    size_type rlen = xstr_min(n, length() - pos);
    memcpy(s, str + pos, rlen);
    return rlen;
}

substring::const_reference
substring::at(size_type pos) const
{
    ASSERT(pos < length());
    return str[pos];
}

substring&
substring::assign(const substring& str, size_type pos, size_type n)
{
    assign(substring(str, pos, n));
    return *this;
}

substring&
substring::assign(const simple_string& str)
{
    assign(substring(str));
    return *this;
}

substring&
substring::assign(const simple_string& str, size_type pos, size_type n)
{
    assign(substring(str, pos, n));
    return *this;
}

substring&
substring::assign(const charT* s)
{
    assign(substring(s));
    return *this;
}

int
substring::compare(const substring& str) const
{
    return cstring_util::compare(data_or_null(), length(),
                                 str.data_or_null(), length());
}

int
substring::compare(size_type pos1, size_type n1, const substring& str) const
{
    return cstring_util::compare(data_or_null() + pos1,
                                 xstr_min(n1, length() - pos1),
                                 str.data_or_null(),
                                 length());
}

int
substring::compare(size_type pos1, size_type n1,
                   const substring& str,
                   size_type pos2, size_type n2) const
{
    return cstring_util::compare(data_or_null() + pos1,
                                 xstr_min(n1, length() - pos1),
                                 str.data_or_null() + pos2,
                                 xstr_min(n2, str.length() - pos2));
}

int
substring::compare(const charT* s) const
{
    return cstring_util::compare(data_or_null(), length(),
                                 s, std::strlen(s));
}

int
substring::compare(size_type pos1, size_type n1,
                   const charT* s, size_type n2) const
{
    return cstring_util::compare(data_or_null() + pos1,
                                 xstr_min(n1, length() - pos1),
                                 s, n2);
}

substring
substring::substr(size_type pos, size_type n) const
{
    return substring(data_or_null() + pos, xstr_min(n, length() - pos));
}

substring::size_type
substring::find (const substring& str, size_type pos) const
{
    return cstring_util::find(data_or_null(), length(), pos,
                              str.data_or_null(), str.length());
}

substring::size_type
substring::find (const charT* s, size_type pos, size_type n) const
{
    return cstring_util::find(data_or_null(), length(), pos, s, n);
}

substring::size_type
substring::find (const charT* s, size_type pos) const
{
    return cstring_util::find(data_or_null(), length(), pos,
                              s, std::strlen(s));
}

substring::size_type
substring::find (charT c, size_type pos) const
{
    return cstring_util::find(data_or_null(), length(), pos, c);
}

substring::size_type
substring::rfind(const substring& str, size_type pos) const
{
    return cstring_util::rfind(data_or_null(), length(), pos,
                               str.data_or_null(), str.length());
}

substring::size_type
substring::rfind(const charT* s, size_type pos, size_type n) const
{
    return cstring_util::rfind(data_or_null(), length(), pos, s, n);
}

substring::size_type
substring::rfind(const charT* s, size_type pos) const
{
    return cstring_util::rfind(data_or_null(), length(), pos,
                               s, std::strlen(s));
}

substring::size_type
substring::rfind(charT c, size_type pos) const
{
    return cstring_util::rfind(data_or_null(), length(), pos, c);
}

substring::size_type
substring::find_first_of(const substring& str, size_type pos) const
{
    return cstring_util::find_first_of(data_or_null(), length(), pos,
                                       str.data_or_null(), str.length());
}

substring::size_type
substring::find_first_of(const charT* s, size_type pos, size_type n) const
{
    return cstring_util::find_first_of(data_or_null(), length(), pos, s, n);
}

substring::size_type
substring::find_first_of(const charT* s, size_type pos) const
{
    return cstring_util::find_first_of(data_or_null(), length(), pos,
                                       s, std::strlen(s));
}

substring::size_type
substring::find_first_of(charT c, size_type pos) const
{
    return cstring_util::find(data_or_null(), length(), pos, c);
}

substring::size_type
substring::find_last_of(const substring& str, size_type pos) const
{
    return cstring_util::find_last_of(data_or_null(), length(), pos,
                                      str.data_or_null(), str.length());
}

substring::size_type
substring::find_last_of(const charT* s, size_type pos, size_type n) const
{
    return cstring_util::find_last_of(data_or_null(), length(), pos,
                                      s, n);
}

substring::size_type
substring::find_last_of(const charT* s, size_type pos) const
{
    return cstring_util::find_last_of(data_or_null(), length(), pos,
                                      s, std::strlen(s));
}

substring::size_type
substring::find_last_of(charT c, size_type pos) const
{
    return cstring_util::rfind(data_or_null(), length(), pos, c);
}

substring::size_type
substring::find_first_not_of(const substring& str, size_type pos) const
{
    return cstring_util::find_first_not_of(data_or_null(), length(), pos,
                                           str.data_or_null(), str.length());
}

substring::size_type
substring::find_first_not_of(const charT* s, size_type pos, size_type n) const
{
    return cstring_util::find_first_not_of(data_or_null(), length(), pos,
                                           s, n);
}

substring::size_type
substring::find_first_not_of(const charT* s, size_type pos) const
{
    return cstring_util::find_first_not_of(data_or_null(), length(), pos,
                                           s, std::strlen(s));
}

substring::size_type
substring::find_first_not_of(charT c, size_type pos) const
{
    return cstring_util::find_first_not_of(data_or_null(), length(), pos,
                                           &c, 1);
}

substring::size_type
substring::find_last_not_of(const substring& str, size_type pos) const
{
    return cstring_util::find_last_not_of(data_or_null(), length(), pos,
                                          str.data_or_null(), length());
}

substring::size_type
substring::find_last_not_of(const charT* s, size_type pos, size_type n) const
{
    return cstring_util::find_last_not_of(data_or_null(), length(), pos, s, n);
}

substring::size_type
substring::find_last_not_of(const charT* s, size_type pos) const
{
    return cstring_util::find_last_not_of(data_or_null(), length(), pos,
                                          s, std::strlen(s));
}

substring::size_type
substring::find_last_not_of(charT c, size_type pos) const
{
    return cstring_util::find_last_not_of(data_or_null(), length(), pos,
                                          &c, 1);
}

substring
strTrim(substring s)
{
    return strRTrim(strLTrim(s));
}

substring
strLTrim(substring s)
{
    substring::size_type n = 0;
    while (n < s.length() && isspace(s[n]))
        ++n;
    return s.substr(n);
}

substring
strRTrim(substring s)
{
    substring::size_type n = s.length();
    while (n > 0 && isspace(s[n-1]))
        --n;
    return s.substr(0, n);
}

substring
strFirst(substring str, substring sep)
{
    substring::size_type n = str.find(sep);
    if (n == substring::npos)
        return str;
    else
        return str.substr(0, n);
}
